class DataTransferObject(object):
    def to_dict(self):
        raise NotImplementedError()

    @classmethod
    def from_dict(cls, payload):
        raise NotImplementedError()


class ImplementsToDto(object):
    def to_dto(self):
        raise NotImplementedError()


class ImplementsFromDto(object):
    @classmethod
    def from_dto(cls, dto_obj):
        raise NotImplementedError()



class PrintableTreeNode(object):
    @property
    def members(self):
        raise NotImplementedError()

    @property
    def detail(self):
        raise NotImplementedError()

    @property
    def is_leaf(self):
        raise NotImplementedError()

    def get_title(self, verbose):
        raise NotImplementedError()


def tree_to_lines(node, verbose=False, title_prefix="", indent=""):
    """
    Return sequence of strings representing lines to print out tree structure on
    command line.
    """
    result = []
    note = ""
    if node.is_leaf:
        note = " [displayed elsewhere]"
    title = node.get_title(verbose)
    result.append("{}{}{}".format(title_prefix, title, note))
    if node.is_leaf:
        return result
    _indent = "|  "
    if not node.members:
        _indent = "   "
    for line in node.detail:
        result.append("{}{}{}".format(indent, _indent, line))
    _indent = "|  "
    _title_prefix = "|- "
    for member in node.members:
        if member == node.members[-1]:
            _indent = "   "
            _title_prefix = "`- "
        result.extend(
            tree_to_lines(
                member,
                verbose,
                indent="{}{}".format(indent, _indent),
                title_prefix="{}{}".format(indent, _title_prefix),
            )
        )
    return result



from __future__ import print_function

from pcs.common.pacemaker.resource.relations import (
    ResourceRelationDto,
    ResourceRelationType,
)
from pcs.cli.common.console_report import format_optional
from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.printable_tree import (
    tree_to_lines,
    PrintableTreeNode,
)


def show_resource_relations_cmd(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file
      * --full - show constraint ids and resource types
    """
    if len(argv) != 1:
        raise CmdLineInputError()
    tree = ResourcePrintableNode.from_dto(
        ResourceRelationDto.from_dict(
            lib.resource.get_resource_relations_tree(argv[0])
        )
    )
    for line in tree_to_lines(tree, verbose=modifiers["full"]):
        print(line)


class ResourceRelationBase(PrintableTreeNode):
    def __init__( self, relation_entity, members, is_leaf):
        self._relation_entity = relation_entity
        self._members = members
        self._is_leaf = is_leaf

    @property
    def is_leaf(self):
        return self._is_leaf

    @property
    def relation_entity(self):
        return self._relation_entity

    @property
    def members(self):
        return self._members

    @property
    def detail(self):
        raise NotImplementedError()

    def get_title(self, verbose):
        raise NotImplementedError()


class ResourcePrintableNode(ResourceRelationBase):
    @classmethod
    def from_dto(cls, resource_dto):
        def _relation_comparator(item):
            type_priorities = (
                ResourceRelationType.INNER_RESOURCES,
                ResourceRelationType.OUTER_RESOURCE,
                ResourceRelationType.ORDER,
                ResourceRelationType.ORDER_SET,
            )
            priority_map = {
                _type: value for value, _type in enumerate(type_priorities)
            }
            return "{_type}_{_id}".format(
                _type=priority_map.get(
                    # Hardcoded number 9 is intentional. If there is more than
                    # 10 items, it would be required to also prepend zeros for
                    # lower numbers. E.g: if there is 100 options, it should
                    # starts as 000, 001, ...
                    item.relation_entity.type, 9 # type: ignore
                ),
                _id=item.relation_entity.id
            )

        return cls(
            resource_dto.relation_entity,
            sorted(
                [
                    RelationPrintableNode.from_dto(member_dto)
                    for member_dto in resource_dto.members
                ],
                key=_relation_comparator,
            ),
            resource_dto.is_leaf
        )

    def get_title(self, verbose):
        rsc_type = self._relation_entity.type
        metadata = self._relation_entity.metadata
        if rsc_type == "primitive":
            rsc_type = "{_class}{_provider}{_type}".format(
                _class=format_optional(metadata.get("class"), "{}:"),
                _provider=format_optional(metadata.get("provider"), "{}:"),
                _type=metadata.get("type"),
            )
        detail = " (resource: {})".format(rsc_type) if verbose else ""
        return "{}{}".format(self._relation_entity.id, detail)

    @property
    def detail(self):
        return []


class RelationPrintableNode(ResourceRelationBase):
    @classmethod
    def from_dto(cls, relation_dto):
        return cls(
            relation_dto.relation_entity,
            sorted(
                [
                    ResourcePrintableNode.from_dto(member_dto)
                    for member_dto in relation_dto.members
                ],
                key=lambda item: item.relation_entity.id,
            ),
            relation_dto.is_leaf
        )

    def get_title(self, verbose):
        rel_type_map = {
            ResourceRelationType.ORDER: "order",
            ResourceRelationType.ORDER_SET: "order set",
            ResourceRelationType.INNER_RESOURCES: "inner resource(s)",
            ResourceRelationType.OUTER_RESOURCE: "outer resource",
        }
        detail = (
            " ({})".format(self._relation_entity.metadata.get("id"))
            if verbose
            else ""
        )
        return "{type}{detail}".format(
            type=rel_type_map.get(self._relation_entity.type, "<unknown>"),
            detail=detail,
        )

    @property
    def detail(self):
        ent = self._relation_entity
        if ent.type == ResourceRelationType.ORDER:
            return _order_metadata_to_str(ent.metadata)
        if ent.type == ResourceRelationType.ORDER_SET:
            return _order_set_metadata_to_str(ent.metadata)
        if (
            ent.type == ResourceRelationType.INNER_RESOURCES
            and
            len(ent.members) > 1
        ):
            return ["members: {}".format(" ".join(ent.members))]
        return []


def _order_metadata_to_str(metadata):
    return [
        "{action1} {resource1} then {action2} {resource2}".format(
            action1=metadata["first-action"],
            resource1=metadata["first"],
            action2=metadata["then-action"],
            resource2=metadata["then"],
        )
    ] + _order_common_metadata_to_str(metadata)


def _order_set_metadata_to_str(metadata):
    result = []
    for res_set in metadata["sets"]:
        result.append("   set {resources}{options}".format(
            resources=" ".join(res_set["members"]),
            options=_resource_set_options_to_str(res_set["metadata"]),
        ))
    return _order_common_metadata_to_str(metadata) + result


def _resource_set_options_to_str(metadata):
    supported_keys = (
        "sequential", "require-all", "ordering", "action", "role", "kind",
        "score",
    )
    result = _filter_supported_keys(metadata, supported_keys)
    return " ({})".format(result) if result else ""


def _filter_supported_keys(data, supported_keys):
    return " ".join([
        "{}={}".format(key, value)
        for key, value in sorted(data.items())
        if key in supported_keys
    ])


def _order_common_metadata_to_str(metadata):
    result = _filter_supported_keys(
        metadata, ("symmetrical", "kind", "require-all", "score")
    )
    return [result] if result else []




from pcs.common.pacemaker.resource.relations import (
    RelationEntityDto,
    ResourceRelationDto,
    ResourceRelationType,
)
from pcs.lib.cib import tools
from pcs.lib.cib.resource import common

from pcs.lib.cib.resource.bundle import TAG as TAG_BUNDLE
from pcs.lib.cib.resource.clone import ALL_TAGS as TAG_CLONE_ALL
from pcs.lib.cib.resource.group import TAG as TAG_GROUP
from pcs.lib.cib.resource.primitive import TAG as TAG_PRIMITIVE


# character ':' ensures that there is no conflict with any id in CIB, as it
# would be an invalid id
INNER_RESOURCE_ID_TEMPLATE = "inner:{}"
OUTER_RESOURCE_ID_TEMPLATE = "outer:{}"


def _get_opposite_relation_id_template(relation_type):
    return {
        ResourceRelationType.INNER_RESOURCES: OUTER_RESOURCE_ID_TEMPLATE,
        ResourceRelationType.OUTER_RESOURCE: INNER_RESOURCE_ID_TEMPLATE,
    }.get(relation_type, "")


class ResourceRelationNode(object):
    def __init__(self, entity):
        self._obj = entity
        self._members = []
        self._is_leaf = False
        self._parent = None
        self._opposite_id = _get_opposite_relation_id_template(
            self._obj.type
        ).format(self._obj.metadata["id"])

    @property
    def obj(self):
        return self._obj

    @property
    def members(self):
        return self._members

    def to_dto(self):
        return ResourceRelationDto(
            self._obj,
            [member.to_dto() for member in self._members],
            self._is_leaf,
        )

    def stop(self):
        self._is_leaf = True

    def add_member(self, member):
        # pylint: disable=protected-access
        if member._parent is not None:
            raise AssertionError(
                "object {} already has a parent set: {}".format(
                    repr(member), repr(member._parent)
                )
            )
        # we don't want opposite relations (inner resource vs outer resource)
        # in a branch, so we are filtering them out
        parents = set(self._get_all_parents())
        if (
            self != member
            and
            member.obj.id not in parents
            and
            (
                member._opposite_id not in parents
                or
                len(member.obj.members) > 1
            )
        ):
            member._parent = self
            self._members.append(member)

    def _get_all_parents(self):
        # pylint: disable=protected-access
        if self._parent is None:
            return []
        return self._parent._get_all_parents() + [self._parent.obj.id]


class ResourceRelationTreeBuilder(object):
    def __init__(self, resource_entities, relation_entities):
        self._resources = resource_entities
        self._all = dict(resource_entities)
        self._all.update(relation_entities)
        self._init_structures()

    def _init_structures(self):
        self._processed_nodes = set()
        # queue
        self._nodes_to_process = []

    def get_tree(self, resource_id):
        self._init_structures()
        if resource_id not in self._resources:
            raise AssertionError(
                "Resource with id '{}' not found in resource "
                "relation structures".format(resource_id)
            )

        # self._all is a superset of self._resources, see __init__
        root = ResourceRelationNode(self._all[resource_id])
        self._nodes_to_process.append(root)

        while self._nodes_to_process:
            node = self._nodes_to_process.pop(0)
            if node.obj.id in self._processed_nodes:
                node.stop()
                continue
            self._processed_nodes.add(node.obj.id)
            for node_id in node.obj.members:
                node.add_member(
                    ResourceRelationNode(self._all[node_id])
                )
            self._nodes_to_process.extend(node.members)
        return root


class ResourceRelationsFetcher(object):
    def __init__(self, cib):
        self._cib = cib
        self._resources_section = tools.get_resources(self._cib)
        self._constraints_section = tools.get_constraints(self._cib)

    def get_relations(self, resource_id):
        resources_to_process = {resource_id}
        relations = {}
        resources = {}
        while resources_to_process:
            res_id = resources_to_process.pop()
            if res_id in resources:
                # already processed
                continue
            res_el = self._get_resource_el(res_id)
            res_relations = {
                rel.id: rel for rel in self._get_resource_relations(res_el)
            }
            resources[res_id] = RelationEntityDto(
                res_id,
                res_el.tag,
                list(res_relations.keys()),
                dict(res_el.attrib),
            )
            relations.update(res_relations)
            resources_to_process.update(
                self._get_all_members(res_relations.values())
            )
        return resources, relations

    def _get_resource_el(self, res_id):
        # client of this class should ensure that res_id really exists in CIB,
        # so here we don't need to handle possible reports
        for tag in TAG_CLONE_ALL + [TAG_GROUP, TAG_PRIMITIVE, TAG_BUNDLE]:
            element_list = self._resources_section.xpath(
                './/{}[@id="{}"]'.format(tag, res_id)
            )
            if element_list:
                return element_list[0]

    @staticmethod
    def _get_all_members(relation_list):
        result = set()
        for relation in relation_list:
            result.update(relation.members)
        return result

    def _get_resource_relations(self, resource_el):
        resource_id = resource_el.attrib["id"]
        relations = [
            _get_ordering_constraint_relation(item)
            for item in self._get_ordering_coinstraints(resource_id)
        ] + [
            _get_ordering_set_constraint_relation(item)
            for item in self._get_ordering_set_constraints(resource_id)
        ]

        # special type of relation, group (note that a group can be a resource
        # and a relation)
        if common.is_wrapper_resource(resource_el):
            relations.append(_get_inner_resources_relation(resource_el))

        # handle resources in a wrapper resource (group/bundle/clone relation)
        parent_el = common.get_parent_resource(resource_el)
        if parent_el is not None:
            relations.append(_get_outer_resource_relation(parent_el))
        return relations

    def _get_ordering_coinstraints(self, resource_id):
        return self._constraints_section.xpath("""
            .//rsc_order[
                not (descendant::resource_set)
                and
                (@first='{0}' or @then='{0}')
            ]
        """.format(resource_id))

    def _get_ordering_set_constraints(self, resource_id):
        return self._constraints_section.xpath(
            ".//rsc_order[./resource_set/resource_ref[@id='{}']]".format(
                resource_id
            )
        )


# relation obj to RelationEntityDto obj
def _get_inner_resources_relation(parent_resource_el):
    attrs = parent_resource_el.attrib
    return RelationEntityDto(
        INNER_RESOURCE_ID_TEMPLATE.format(attrs["id"]),
        ResourceRelationType.INNER_RESOURCES,
        [
            res.attrib["id"]
            for res in common.get_inner_resources(parent_resource_el)
        ],
        dict(attrs),
    )


def _get_outer_resource_relation(parent_resource_el):
    attrs = parent_resource_el.attrib
    return RelationEntityDto(
        OUTER_RESOURCE_ID_TEMPLATE.format(attrs["id"]),
        ResourceRelationType.OUTER_RESOURCE,
        [attrs["id"]],
        dict(attrs),
    )


def _get_ordering_constraint_relation(ord_const_el):
    attrs = ord_const_el.attrib
    return RelationEntityDto(
        attrs["id"],
        ResourceRelationType.ORDER,
        [attrs["first"], attrs["then"]],
        dict(attrs),
    )


def _get_ordering_set_constraint_relation(ord_set_const_el):
    attrs = ord_set_const_el.attrib
    members = set()
    metadata = dict(attrs)
    metadata["sets"] = []
    for rsc_set_el in ord_set_const_el.findall("resource_set"):
        rsc_set = dict(
            id=rsc_set_el.get("id"),
            metadata=dict(rsc_set_el.attrib),
            members=[],
        )
        metadata["sets"].append(rsc_set)
        for rsc_ref in rsc_set_el.findall("resource_ref"):
            rsc_id = rsc_ref.attrib["id"]
            members.add(rsc_id)
            rsc_set["members"].append(rsc_id)

    return RelationEntityDto(
        attrs["id"], ResourceRelationType.ORDER_SET, sorted(members), metadata
    )



from pcs.common.interface.dto import DataTransferObject


class ResourceRelationType(object):
    ORDER = "ORDER"
    ORDER_SET = "ORDER_SET"
    INNER_RESOURCES = "INNER_RESOURCES"
    OUTER_RESOURCE = "OUTER_RESOURCE"


class RelationEntityDto(DataTransferObject):
    def __init__(self, id_, type_, members, metadata):
        # pylint: disable=invalid-name
        self.id = id_
        self.type = type_
        self.members = members
        self.metadata = metadata

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and
            self.to_dict() == other.to_dict()
        )

    def to_dict(self):
        return dict(
            id=self.id,
            type=self.type,
            members=self.members,
            metadata=self.metadata,
        )

    @classmethod
    def from_dict(cls, payload):
        return cls(
            payload["id"],
            payload["type"],
            payload["members"],
            payload["metadata"],
        )


class ResourceRelationDto(DataTransferObject):
    def __init__(self, relation_entity, members, is_leaf):
        self.relation_entity = relation_entity
        self.members = members
        self.is_leaf = is_leaf

    def to_dict(self):
        return dict(
            relation_entity=self.relation_entity.to_dict(),
            members=[member.to_dict() for member in self.members],
            is_leaf=self.is_leaf,
        )

    @classmethod
    def from_dict(cls, payload):
        return cls(
            RelationEntityDto.from_dict(payload["relation_entity"]),
            [
                ResourceRelationDto.from_dict(member_data)
                for member_data in payload["members"]
            ],
            payload["is_leaf"],
        )



from __future__ import (
    absolute_import,
    division,
    print_function,
)

import getopt
import os
import sys
import logging

from pcs import (
    acl,
    alert,
    booth,
    client,
    cluster,
    config,
    constraint,
    node,
    pcsd,
    prop,
    qdevice,
    quorum,
    resource,
    settings,
    status,
    stonith,
    usage,
    utils,
)

from pcs.cli.common import (
    capabilities,
    completion,
    errors,
    parse_args,
)
from pcs.lib.errors import LibraryError


logging.basicConfig()
usefile = False
filename = ""

def _non_root_run(argv_cmd):
    """
    This function will run commands which has to be run as root for users which
    are not root. If it required to run such command as root it will do that by
    sending it to the local pcsd and then it will exit.
    """
    # matching the commands both in here and in pcsd expects -o and --options
    # to be at the end of a command
    argv_and_options = argv_cmd[:]
    for option, value in utils.pcs_options.items():
        if parse_args.is_option_expecting_value(option):
            argv_and_options.extend([option, value])
        else:
            argv_and_options.append(option)

    # specific commands need to be run under root account, pass them to pcsd
    # don't forget to allow each command in pcsd.rb in "post /run_pcs do"
    root_command_list = [
        ['cluster', 'auth', '...'],
        ['cluster', 'corosync', '...'],
        ['cluster', 'destroy', '...'],
        ['cluster', 'disable', '...'],
        ['cluster', 'enable', '...'],
        ['cluster', 'node', '...'],
        ['cluster', 'pcsd-status', '...'],
        ['cluster', 'setup', '...'],
        ['cluster', 'start', '...'],
        ['cluster', 'stop', '...'],
        ['cluster', 'sync', '...'],
        # ['config', 'restore', '...'], # handled in config.config_restore
        ['pcsd', 'sync-certificates'],
        ["quorum", "device", "status", "..."],
        ["quorum", "status", "..."],
        ['status', 'corosync', '...'],
        ['status', 'nodes', 'corosync-id'],
        ['status', 'nodes', 'pacemaker-id'],
        ['status', 'pcsd', '...'],
        ['status', 'quorum', '...'],
    ]
    for root_cmd in root_command_list:
        if (
            (argv_and_options == root_cmd)
            or
            (
                root_cmd[-1] == "..."
                and
                argv_and_options[:len(root_cmd)-1] == root_cmd[:-1]
            )
        ):
            # handle interactivity of 'pcs cluster auth'
            if argv_and_options[0:2] == ["cluster", "auth"]:
                if "-u" not in utils.pcs_options:
                    username = utils.get_terminal_input('Username: ')
                    argv_and_options.extend(["-u", username])
                if "-p" not in utils.pcs_options:
                    password = utils.get_terminal_password()
                    argv_and_options.extend(["-p", password])

            # call the local pcsd
            err_msgs, exitcode, std_out, std_err = utils.call_local_pcsd(
                argv_and_options
            )
            if err_msgs:
                for msg in err_msgs:
                    utils.err(msg, False)
                sys.exit(1)
            if std_out.strip():
                print(std_out)
            if std_err.strip():
                sys.stderr.write(std_err)
            sys.exit(exitcode)

def main(argv=None):
    if completion.has_applicable_environment(os.environ):
        print(completion.make_suggestions(
            os.environ,
            usage.generate_completion_tree_from_usage()
        ))
        sys.exit()

    argv = argv if argv else sys.argv[1:]
    utils.subprocess_setup()
    global filename, usefile
    utils.pcs_options = {}

    argv = parse_args.upgrade_args(argv)

    # we want to support optional arguments for --wait, so if an argument
    # is specified with --wait (ie. --wait=30) then we use them
    waitsecs = None
    new_argv = []
    for arg in argv:
        if arg.startswith("--wait="):
            tempsecs = arg.replace("--wait=","")
            if len(tempsecs) > 0:
                waitsecs = tempsecs
                arg = "--wait"
        new_argv.append(arg)
    argv = new_argv

    try:
        pcs_options, dummy_argv = getopt.gnu_getopt(
            parse_args.filter_out_non_option_negative_numbers(argv),
            parse_args.PCS_SHORT_OPTIONS,
            parse_args.PCS_LONG_OPTIONS,
        )
    except getopt.GetoptError as err:
        print(err)
        usage.main()
        sys.exit(1)
    argv = parse_args.filter_out_options(argv)

    full = False
    for option, dummy_value in pcs_options:
        if option == "--full":
            full = True
            break

    for o, a in pcs_options:
        if not o in utils.pcs_options:
            if o in ["--watchdog", "--device"]:
                a = [a]
            utils.pcs_options[o] = a
        else:
            # If any options are a list then they've been entered twice which isn't valid
            if o not in ["--watchdog", "--device"]:
                utils.err("%s can only be used once" % o)
            else:
                utils.pcs_options[o].append(a)

        if o == "-h" or o == "--help":
            if len(argv) == 0:
                usage.main()
                sys.exit()
            else:
                argv = [argv[0], "help" ] + argv[1:]
        elif o == "-f":
            usefile = True
            filename = a
            utils.usefile = usefile
            utils.filename = filename
        elif o == "--corosync_conf":
            settings.corosync_conf_file = a
        elif o == "--cluster_conf":
            settings.cluster_conf_file = a
        elif o == "--version":
            print(settings.pcs_version)
            if full:
                print(" ".join(
                    sorted([
                        feat["id"]
                        for feat in capabilities.get_pcs_capabilities()
                    ])
                ))
            sys.exit()
        elif o == "--fullhelp":
            usage.full_usage()
            sys.exit()
        elif o == "--wait":
            utils.pcs_options[o] = waitsecs
        elif o == "--request-timeout":
            request_timeout_valid = False
            try:
                timeout = int(a)
                if timeout > 0:
                    utils.pcs_options[o] = timeout
                    request_timeout_valid = True
            except ValueError:
                pass
            if not request_timeout_valid:
                utils.err(
                    (
                        "'{0}' is not a valid --request-timeout value, use "
                        "a positive integer"
                    ).format(a)
                )

    if len(argv) == 0:
        usage.main()
        sys.exit(1)

    # create a dummy logger
    # we do not have a log file for cli (yet), but library requires a logger
    logger = logging.getLogger("pcs")
    logger.propagate = 0
    logger.handlers = []

    # root can run everything directly, also help can be displayed,
    # working on a local file also do not need to run under root
    if (os.getuid() != 0) and (argv and argv[0] != "help") and not usefile:
        _non_root_run(argv)

    command = argv.pop(0)
    if (command == "-h" or command == "help"):
        usage.main()
        return
    cmd_map = {
        "resource": resource.resource_cmd,
        "cluster": cluster.cluster_cmd,
        "stonith": stonith.stonith_cmd,
        "property": prop.property_cmd,
        "constraint": constraint.constraint_cmd,
        "acl": lambda argv: acl.acl_cmd(
            utils.get_library_wrapper(),
            argv,
            utils.get_modifiers()
        ),
        "status": lambda argv: status.status_cmd(
            utils.get_library_wrapper(),
            argv,
            utils.get_modifiers()
        ),
        "config": config.config_cmd,
        "pcsd": pcsd.pcsd_cmd,
        "node": lambda argv: node.node_cmd(
            utils.get_library_wrapper(),
            argv,
            utils.get_modifiers()
        ),
        "quorum": lambda argv: quorum.quorum_cmd(
            utils.get_library_wrapper(),
            argv,
            utils.get_modifiers()
        ),
        "qdevice": lambda argv: qdevice.qdevice_cmd(
            utils.get_library_wrapper(),
            argv,
            utils.get_modifiers()
        ),
        "alert": lambda args: alert.alert_cmd(
            utils.get_library_wrapper(),
            args,
            utils.get_modifiers()
        ),
        "booth": lambda argv: booth.booth_cmd(
            utils.get_library_wrapper(),
            argv,
            utils.get_modifiers()
        ),
        "client": lambda argv: client.client_cmd(
            utils.get_library_wrapper(),
            argv,
            utils.get_modifiers()
        ),
    }
    if command not in cmd_map:
        usage.main()
        sys.exit(1)

    try:
        cmd_map[command](argv)
    except LibraryError as e:
        utils.process_library_reports(e.args)
    except errors.CmdLineInputError:
        argv = [command] + argv
        if argv and argv[0] in cmd_map:
            usage.show(argv[0], argv[1:])
        else:
            usage.main()
        sys.exit(1)
